use std::sync::Arc;

use bevy_ecs::component::Component;
use glam::{Mat4, Vec3, Vec4};
use vulkano::{
    Validated,
    buffer::{AllocateBufferError, Subbuffer},
    memory::allocator::StandardMemoryAllocator,
};

use super::{AsUniformBuffer, mvp::Mvp};

// See docs/OPENGL_VULKAN_DIFF.md
const OPENGL_TO_VULKAN_Y_AXIS_FLIP: Mat4 = Mat4 {
    x_axis: Vec4::new(1.0, 0.0, 0.0, 0.0),
    y_axis: Vec4::new(0.0, -1.0, 0.0, 0.0),
    z_axis: Vec4::new(0.0, 0.0, 1.0, 0.0),
    w_axis: Vec4::new(0.0, 0.0, 0.0, 1.0),
};

#[derive(Component)]
pub struct Camera3D {
    projection: Mat4,
    aspect_ratio: f32,
    fov: f32,
    near: f32,
    far: f32,
}

#[derive(Component)]
pub struct Camera3DTransform {
    pub position: Vec3,
    pub rotation: Vec3,
}

impl Camera3D {
    pub fn new(aspect_ratio: f32, fov: f32, near: f32, far: f32) -> Self {
        Self {
            projection: Mat4::perspective_rh(fov, aspect_ratio, near, far),
            aspect_ratio,
            fov,
            near,
            far,
        }
    }

    pub fn update_projection(&mut self, window_aspect_ratio: f32) {
        if self.aspect_ratio != window_aspect_ratio {
            self.aspect_ratio = window_aspect_ratio;
            self.projection =
                Mat4::perspective_rh(self.fov, self.aspect_ratio, self.near, self.far);
        }
    }

    pub fn set_projection(&mut self, projection: Mat4) {
        self.projection = projection;
    }

    pub fn create_buffer(
        &self,
        transform: &Camera3DTransform,
        memory_allocator: &Arc<StandardMemoryAllocator>,
    ) -> Result<Subbuffer<[Mvp]>, Validated<AllocateBufferError>> {
        let (sin_pitch, cos_pitch) = transform.rotation.x.sin_cos();
        let (sin_yaw, cos_yaw) = transform.rotation.y.sin_cos();

        let view_matrix = Mat4::look_to_rh(
            transform.position,
            Vec3::new(cos_pitch * cos_yaw, sin_pitch, cos_pitch * sin_yaw).normalize(),
            Vec3::Y,
        );

        let mvp = Mvp {
            model: OPENGL_TO_VULKAN_Y_AXIS_FLIP.to_cols_array_2d(),
            view: view_matrix.to_cols_array_2d(),
            projection: self.projection.to_cols_array_2d(),
        };

        mvp.create_uniform_buffer(memory_allocator)
    }
}
